# Copyright 2016 Canonical Limited.  All rights reserved.

# To run: "python -m unittest test_collect-logs"

import errno
import os
import os.path
import shutil
import subprocess
import sys
import tempfile
from unittest import TestCase

import mock


__file__ = os.path.abspath(__file__)

script = type(sys)("collect-logs")
script.__file__ = os.path.abspath("collect-logs")
execfile("collect-logs", script.__dict__)


class FakeError(Exception):
    """A specific error for which to check."""


def _touch(filename):
     with open(filename, "w"):
         pass


class _BaseTestCase(TestCase):

    MOCKED = None

    def setUp(self):
        super(_BaseTestCase, self).setUp()

        self.orig_cwd = os.getcwd()
        self.cwd = tempfile.mkdtemp()
        os.chdir(self.cwd)

        self.tempdir = os.path.join(self.cwd, "tempdir")
        os.mkdir(self.tempdir)

        self.orig = {}
        for attr in self.MOCKED or ():
            self.orig[attr] = getattr(script, attr)
            setattr(script, attr, mock.Mock())

    def tearDown(self):
        for attr in self.MOCKED or ():
            setattr(script, attr, self.orig[attr])

        shutil.rmtree(self.cwd)
        os.chdir(self.orig_cwd)

        super(_BaseTestCase, self).tearDown()

    def _touch_tempfile(self, filename):
        try:
            os.makedirs(os.path.join(self.tempdir, os.path.dirname(filename)))
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
        _touch(os.path.join(self.tempdir, filename))

    def assert_cwd(self, dirname):
        cwd = os.getcwd()
        self.assertEqual(cwd, dirname)


class MainTestCase(_BaseTestCase):

    MOCKED = ("collect_logs", "collect_inner_logs", "bundle_logs")

    def setUp(self):
        super(MainTestCase, self).setUp()

        self.orig_mkdtemp = script.mkdtemp
        script.mkdtemp = lambda: self.tempdir

    def tearDown(self):
        script.mkdtemp = self.orig_mkdtemp

        super(MainTestCase, self).tearDown()

    def test_success(self):
        tarfile = "/tmp/logs.tgz"
        extrafiles = ["spam.py"]

        script.main(tarfile, extrafiles)

        script.collect_logs.assert_called_once_with()
        script.collect_inner_logs.assert_called_once_with()
        script.bundle_logs.assert_called_once_with(
            self.tempdir, tarfile, extrafiles)
        self.assertFalse(os.path.exists(self.tempdir))

    def test_in_correct_directories(self):
        script.collect_logs.side_effect = lambda: self.assert_cwd(self.tempdir)
        script.collect_inner_logs.side_effect = lambda: self.assert_cwd(self.tempdir)
        script.bundle_logs.side_effect = lambda *a: self.assert_cwd(self.cwd)
        tarfile = "/tmp/logs.tgz"
        extrafiles = ["spam.py"]

        script.main(tarfile, extrafiles)

    def test_cleanup(self):
        tarfile = "/tmp/logs.tgz"
        extrafiles = ["spam.py"]

        script.main(tarfile, extrafiles)

        self.assertFalse(os.path.exists(self.tempdir))

    def test_collect_logs_error(self):
        tarfile = "/tmp/logs.tgz"
        extrafiles = ["spam.py"]
        script.collect_logs.side_effect = FakeError()

        with self.assertRaises(FakeError):
            script.main(tarfile, extrafiles)

        script.collect_logs.assert_called_once_with()
        script.collect_inner_logs.assert_not_called()
        script.bundle_logs.assert_not_called()
        self.assertFalse(os.path.exists(self.tempdir))

    def test_collect_inner_logs_error(self):
        tarfile = "/tmp/logs.tgz"
        extrafiles = ["spam.py"]
        script.collect_inner_logs.side_effect = FakeError()

        script.main(tarfile, extrafiles)

        script.collect_logs.assert_called_once_with()
        script.collect_inner_logs.assert_called_once_with()
        script.bundle_logs.assert_called_once_with(
            self.tempdir, tarfile, extrafiles)
        self.assertFalse(os.path.exists(self.tempdir))

    def test_bundle_logs_error(self):
        tarfile = "/tmp/logs.tgz"
        extrafiles = ["spam.py"]
        script.bundle_logs.side_effect = FakeError()

        with self.assertRaises(FakeError):
            script.main(tarfile, extrafiles)

        script.collect_logs.assert_called_once_with()
        script.collect_inner_logs.assert_called_once_with()
        script.bundle_logs.assert_called_once_with(
            self.tempdir, tarfile, extrafiles)
        self.assertFalse(os.path.exists(self.tempdir))


class CollectLogsTestCase(_BaseTestCase):

    MOCKED = ("get_units", "check_output", "call")

    def setUp(self):
        super(CollectLogsTestCase, self).setUp()

        self.units = [
            "landscape-server/0",
            "postgresql/0",
            "rabbitmq-server/0",
            "haproxy/0",
            ]
        script.get_units.return_value = self.units[:]

        self.mp_map_orig = script._mp_map
        script._mp_map = lambda f, a: map(f, a)

        os.chdir(self.tempdir)

    def tearDown(self):
        script._mp_map = self.mp_map_orig

        super(CollectLogsTestCase, self).tearDown()

    def _call_side_effect(self, cmd):
        if cmd[0] != "tar":
            self.assertTrue(os.path.exists(cmd[-1]))
            return
        self.assertEqual(cmd[0], "juju")
        _touch(os.path.basename(cmd[2]))

    def test_success(self):
        script.call.side_effect = self._call_side_effect

        script.collect_logs()

        script.get_units.assert_called_once_with()
        expected = []
        units = self.units + ["0"]
        # for _create_ps_output_file()
        for unit in units:
            cmd = "ps fauxww | sudo tee /var/log/ps-fauxww.txt"
            expected.append(mock.call(["juju", "ssh", unit, cmd],
                                      stderr=subprocess.STDOUT,
                                      ))
        # for _create_log_tarball()
        for unit in units:
            cmd = ("sudo tar --ignore-failed-read"
                   " --exclude=/var/lib/landscape/client/package/hash-id"
                   " --exclude=/var/lib/juju/containers/juju-*-lxc-template"
                   " -czf /tmp/logs_{}.tar.gz"
                   " $(sudo sh -c \"ls -1d {} 2>/dev/null\")"
                   ).format(
                       unit.replace("/", "-") if unit != "0" else "bootstrap",
                       " ".join(["/var/log",
                                 "/etc/nova",
                                 "/etc/swift",
                                 "/etc/neutron",
                                 "/etc/apache2",
                                 "/etc/haproxy",
                                 "/etc/ceph",
                                 "/etc/glance",
                                 "/var/lib/lxc/*/rootfs/var/log",
                                 "/var/lib/juju/containers",
                                 "/var/lib/landscape/client",
                                 ]),
                       )
            expected.append(mock.call(["juju", "ssh", unit, cmd],
                                      stderr=subprocess.STDOUT,
                                      ))
        self.assertEqual(script.check_output.call_count, len(expected))
        script.check_output.assert_has_calls(expected, any_order=True)
        expected = []
        # for download_log_from_unit()
        for unit in units:
            name = unit.replace("/", "-") if unit != "0" else "bootstrap"
            filename = "logs_{}.tar.gz".format(name)
            source = "{}:/tmp/{}".format(unit, filename)
            expected.append(mock.call(["juju", "scp", source, "."]))
            expected.append(mock.call(["tar", "-C", name, "-xzf", filename]))
            self.assertFalse(os.path.exists(filename))
        self.assertEqual(script.call.call_count, len(expected))
        script.call.assert_has_calls(expected, any_order=True)

    def test_get_units_failure(self):
        script.get_units.side_effect = FakeError()

        with self.assertRaises(FakeError):
            script.collect_logs()

        script.get_units.assert_called_once_with()
        script.check_output.assert_not_called()
        script.call.assert_not_called()

    def test_check_output_failure(self):
        script.check_output.side_effect = [mock.DEFAULT,
                                           FakeError(),
                                           ]

        with self.assertRaises(FakeError):
            script.collect_logs()

        script.get_units.assert_called_once_with()
        self.assertEqual(script.check_output.call_count, 2)
        script.call.assert_not_called()

    def test_call_failure(self):
        def call_side_effect(cmd):
            # second use of call() for landscape-server/0
            if script.call.call_count == 2:
                raise FakeError()
            # first use of call() for postgresql/0
            if script.call.call_count == 3:
                raise FakeError()
            # all other uses of call() default to the normal side effect.
            return self._call_side_effect(cmd)
        script.call.side_effect = call_side_effect

        script.collect_logs()

        script.get_units.assert_called_once_with()
        units = self.units + ["0"]
        self.assertEqual(script.check_output.call_count, len(units)*2)
        self.assertEqual(script.call.call_count, len(units)*2-1)
        for unit in units:
            name = unit.replace("/", "-") if unit != "0" else "bootstrap"
            if unit == self.units[1]:
                self.assertFalse(os.path.exists(name))
            else:
                self.assertTrue(os.path.exists(name))
            filename = "logs_{}.tar.gz".format(name)
            self.assertFalse(os.path.exists(filename))


class CollectInnerLogsTestCase(_BaseTestCase):

    MOCKED = ("get_units", "check_output", "call", "check_call")

    def setUp(self):
        super(CollectInnerLogsTestCase, self).setUp()

        self.units = [
            "landscape-server/0",
            "postgresql/0",
            "rabbitmq-server/0",
            "haproxy/0",
            ]
        script.get_units.return_value = self.units[:]

        os.chdir(self.tempdir)

    def assert_clean(self):
        self.assert_cwd(self.tempdir)
        self.assertFalse(os.path.exists("inner-logs.tar.gz"))

    def test_success(self):
        def check_call_side_effect(cmd):
            if script.check_call.call_count == 2:
                self.assert_cwd(self.tempdir)
                self._touch_tempfile("inner-logs.tar.gz")
            elif script.check_call.call_count == 3:
                cwd = os.path.join(self.tempdir, "landscape-0-inner-logs")
                self.assert_cwd(cwd)
            return None
        script.check_call.side_effect = check_call_side_effect

        script.collect_inner_logs()

        # get_units()
        script.get_units.assert_called_once_with()
        # check_output()
        cmd = ("sudo JUJU_HOME=/var/lib/landscape/juju-homes/"
               "`sudo ls -rt /var/lib/landscape/juju-homes/ | tail -1`"
               " juju status")
        script.check_output.assert_called_once_with(
            ["juju", "ssh", "landscape-server/0", cmd])
        # call()
        expected = [
            mock.call(["juju", "scp",
                       os.path.join(os.path.dirname(__file__), "collect-logs"),
                       "landscape-server/0:/tmp/collect-logs",
                       ]),
            mock.call(["juju", "ssh",
                       "landscape-server/0",
                       "sudo rm -rf /tmp/inner-logs.tar.*",
                       ]),
            ]
        self.assertEqual(script.call.call_count, len(expected))
        script.call.assert_has_calls(expected, any_order=True)
        # check_call()
        cmd = ("sudo -u landscape"
               " JUJU_HOME=/var/lib/landscape/juju-homes/"
               "`sudo ls -rt /var/lib/landscape/juju-homes/ | tail -1`"
               " /tmp/collect-logs /tmp/inner-logs.tar.gz")
        expected = [
            mock.call(["juju", "ssh", "landscape-server/0", cmd]),
            mock.call(["juju", "scp",
                       "landscape-server/0:/tmp/inner-logs.tar.gz",
                       ".",
                       ]),
            mock.call(["tar", "-zxf", self.tempdir+"/inner-logs.tar.gz"]),
            ]
        self.assertEqual(script.check_call.call_count, len(expected))
        script.check_call.assert_has_calls(expected, any_order=True)
        self.assert_clean()

    def test_with_legacy_landscape_unit(self):
        self.units[0] = "landscape/0"
        script.get_units.return_value = self.units[:]

        script.collect_inner_logs()

        cmd = ("sudo JUJU_HOME=/var/lib/landscape/juju-homes/"
               "`sudo ls -rt /var/lib/landscape/juju-homes/ | tail -1`"
               " juju status")
        script.check_output.assert_called_once_with(
            ["juju", "ssh", "landscape/0", cmd])
        self.assert_clean()

    def test_no_units(self):
        script.get_units.return_value = []

        script.collect_inner_logs()

        script.get_units.assert_called_once_with()
        script.check_output.assert_not_called()
        script.call.assert_not_called()
        script.check_call.assert_not_called()
        self.assert_clean()

    def test_no_landscape_server_unit(self):
        del self.units[0]
        script.get_units.return_value = self.units[:]

        script.collect_inner_logs()

        script.get_units.assert_called_once_with()
        script.check_output.assert_not_called()
        script.call.assert_not_called()
        script.check_call.assert_not_called()
        self.assert_clean()

    def test_get_units_failure(self):
        script.get_units.side_effect = FakeError()

        with self.assertRaises(FakeError):
            script.collect_inner_logs()

        self.assertEqual(script.get_units.call_count, 1)
        script.check_output.assert_not_called()
        script.call.assert_not_called()
        script.check_call.assert_not_called()
        self.assert_cwd(self.tempdir)
        self.assert_clean()

    def test_check_output_failure(self):
        script.check_output.side_effect = FakeError()

        with self.assertRaises(FakeError):
            script.collect_inner_logs()

        self.assertEqual(script.get_units.call_count, 1)
        self.assertEqual(script.check_output.call_count, 1)
        script.call.assert_not_called()
        script.check_call.assert_not_called()
        self.assert_cwd(self.tempdir)
        self.assert_clean()

    def test_call_failure_1(self):
        script.call.side_effect = FakeError()

        with self.assertRaises(FakeError):
            script.collect_inner_logs()

        self.assertEqual(script.get_units.call_count, 1)
        self.assertEqual(script.check_output.call_count, 1)
        self.assertEqual(script.call.call_count, 1)
        script.check_call.assert_not_called()
        self.assert_cwd(self.tempdir)
        self.assert_clean()

    def test_call_failure_2(self):
        script.call.side_effect = [None,
                                   FakeError(),
                                   ]

        with self.assertRaises(FakeError):
            script.collect_inner_logs()

        self.assertEqual(script.get_units.call_count, 1)
        self.assertEqual(script.check_output.call_count, 1)
        self.assertEqual(script.call.call_count, 2)
        script.check_call.assert_not_called()
        self.assert_clean()

    def test_check_call_failure_1(self):
        script.check_call.side_effect = FakeError()

        with self.assertRaises(FakeError):
            script.collect_inner_logs()

        self.assertEqual(script.get_units.call_count, 1)
        self.assertEqual(script.check_output.call_count, 1)
        self.assertEqual(script.call.call_count, 2)
        self.assertEqual(script.check_call.call_count, 1)
        self.assert_clean()

    def test_check_call_failure_2(self):
        script.check_call.side_effect = [None,
                                         FakeError(),
                                         ]

        with self.assertRaises(FakeError):
            script.collect_inner_logs()

        self.assertEqual(script.get_units.call_count, 1)
        self.assertEqual(script.check_output.call_count, 1)
        self.assertEqual(script.call.call_count, 2)
        self.assertEqual(script.check_call.call_count, 2)
        self.assert_clean()

    def test_check_call_failure_3(self):
        def check_call_side_effect(cmd):
            if script.check_call.call_count == 1:
                return None
            if script.check_call.call_count == 2:
                self._touch_tempfile("inner-logs.tar.gz")
                return None
            raise FakeError()
        script.check_call.side_effect = check_call_side_effect

        with self.assertRaises(FakeError):
            script.collect_inner_logs()

        self.assertEqual(script.get_units.call_count, 1)
        self.assertEqual(script.check_output.call_count, 1)
        self.assertEqual(script.call.call_count, 2)
        self.assertEqual(script.check_call.call_count, 3)
        self.assert_clean()


class BundleLogsTestCase(_BaseTestCase):

    MOCKED = ("call",)

    def setUp(self):
        super(BundleLogsTestCase, self).setUp()

        os.chdir(self.tempdir)

        self._touch_tempfile("bootstrap/var/log/syslog")
        self._touch_tempfile("bootstrap/var/log/juju/all-machines.log")
        self._touch_tempfile(
            "bootstrap/var/lib/lxc/deadbeef/rootfs/var/log/syslog")
        self._touch_tempfile("bootstrap/var/lib/juju/containers")
        self._touch_tempfile("landscape-server-0/var/log/syslog")
        self._touch_tempfile("postgresql-0/var/log/syslog")
        self._touch_tempfile("rabbitmq-server-0/var/log/syslog")
        self._touch_tempfile("haproxy-0/var/log/syslog")
        self._touch_tempfile("landscape-0-inner-logs/bootstrap/var/log/syslog")

        self.extrafile = os.path.join(self.cwd, "spam.txt")
        _touch(self.extrafile)

    def test_success_with_extra(self):
        tarfile = "/tmp/logs.tgz"
        extrafiles = [self.extrafile]

        script.bundle_logs(self.tempdir, tarfile, extrafiles)

        script.call.assert_called_once_with(
            ["tar",
             "czf", tarfile,
             "--transform", "s,{}/,,".format(self.tempdir[1:]),
             os.path.join(self.tempdir, "bootstrap"),
             os.path.join(self.tempdir, "haproxy-0"),
             os.path.join(self.tempdir, "landscape-0-inner-logs"),
             os.path.join(self.tempdir, "landscape-server-0"),
             os.path.join(self.tempdir, "postgresql-0"),
             os.path.join(self.tempdir, "rabbitmq-server-0"),
             self.extrafile,
             ],
            )

    def test_success_without_extra(self):
        tarfile = "/tmp/logs.tgz"

        script.bundle_logs(self.tempdir, tarfile)

        script.call.assert_called_once_with(
            ["tar",
             "czf", tarfile,
             "--transform", "s,{}/,,".format(self.tempdir[1:]),
             os.path.join(self.tempdir, "bootstrap"),
             os.path.join(self.tempdir, "haproxy-0"),
             os.path.join(self.tempdir, "landscape-0-inner-logs"),
             os.path.join(self.tempdir, "landscape-server-0"),
             os.path.join(self.tempdir, "postgresql-0"),
             os.path.join(self.tempdir, "rabbitmq-server-0"),
             ],
            )

    def test_success_no_files(self):
        for filename in os.listdir(self.tempdir):
            shutil.rmtree(os.path.join(self.tempdir, filename))
        tarfile = "/tmp/logs.tgz"

        script.bundle_logs(self.tempdir, tarfile)

        script.call.assert_called_once_with(
            ["tar",
             "czf", tarfile,
             "--transform", "s,{}/,,".format(self.tempdir[1:]),
             ],
            )

    def test_call_failure(self):
        script.call.side_effect = FakeError()
        tarfile = "/tmp/logs.tgz"

        with self.assertRaises(FakeError):
            script.bundle_logs(self.tempdir, tarfile)

        script.call.assert_called_once_with(
            ["tar",
             "czf", tarfile,
             "--transform", "s,{}/,,".format(self.tempdir[1:]),
             os.path.join(self.tempdir, "bootstrap"),
             os.path.join(self.tempdir, "haproxy-0"),
             os.path.join(self.tempdir, "landscape-0-inner-logs"),
             os.path.join(self.tempdir, "landscape-server-0"),
             os.path.join(self.tempdir, "postgresql-0"),
             os.path.join(self.tempdir, "rabbitmq-server-0"),
             ],
            )
